@@ -1,6 +1,5 @@
-from __future__ import annotations
-
-from typing import Any, Dict, List, Optional, Tuple, Union
+import functools
+from typing import Any
 
 import attrs
 import numpy as np
@@ -12,7 +11,7 @@ import sarsen
 try:
     import dask  # noqa: F401
 
-    DEFAULT_MEASUREMENT_CHUNKS: Optional[int] = 2048
+    DEFAULT_MEASUREMENT_CHUNKS: int | None = 2048
 except ModuleNotFoundError:
     DEFAULT_MEASUREMENT_CHUNKS = None
 
@@ -21,11 +20,11 @@ SPEED_OF_LIGHT = 299_792_458.0  # m / s
 
 def open_dataset_autodetect(
     product_urlpath: str,
-    group: Optional[str] = None,
-    chunks: Optional[Union[int, Dict[str, int]]] = None,
+    group: str | None = None,
+    chunks: int | dict[str, int] | None = None,
     check_files_exist: bool = False,
     **kwargs: Any,
-) -> Tuple[xr.Dataset, Dict[str, Any]]:
+) -> tuple[xr.Dataset, dict[str, Any]]:
     kwargs.setdefault("engine", "sentinel-1")
     try:
         ds = xr.open_dataset(
@@ -37,17 +36,17 @@ def open_dataset_autodetect(
         )
     except FileNotFoundError:
         # re-try with Planetary Computer option
-        kwargs[
-            "override_product_files"
-        ] = "{dirname}/{prefix}{swath}-{polarization}{ext}"
+        kwargs["override_product_files"] = (
+            "{dirname}/{prefix}{swath}-{polarization}{ext}"
+        )
         ds = xr.open_dataset(product_urlpath, group=group, chunks=chunks, **kwargs)
     return ds, kwargs
 
 
 def azimuth_slant_range_grid(
-    attrs: Dict[str, Any],
-    grouping_area_factor: Tuple[float, float] = (3.0, 3.0),
-) -> Dict[str, Any]:
+    attrs: dict[str, Any],
+    grouping_area_factor: tuple[float, float] = (3.0, 3.0),
+) -> dict[str, Any]:
     if attrs["product_type"] == "SLC":
         slant_range_spacing_m = (
             attrs["range_pixel_spacing"]
@@ -61,7 +60,7 @@ def azimuth_slant_range_grid(
         slant_range_spacing_m * 2 / SPEED_OF_LIGHT  # ignore type
     )
 
-    grid_parameters: Dict[str, Any] = {
+    grid_parameters: dict[str, Any] = {
         "slant_range_time0": attrs["image_slant_range_time"],
         "slant_range_time_interval_s": slant_range_time_interval_s,
         "slant_range_spacing_m": slant_range_spacing_m,
@@ -76,11 +75,11 @@ def azimuth_slant_range_grid(
 @attrs.define(slots=False)
 class Sentinel1SarProduct(sarsen.GroundRangeSarProduct, sarsen.SlantRangeSarProduct):
     product_urlpath: str
-    measurement_group: Optional[str] = None
-    measurement_chunks: Optional[int] = DEFAULT_MEASUREMENT_CHUNKS
-    kwargs: Dict[str, Any] = {}
+    measurement_group: str | None = None
+    measurement_chunks: int | None = DEFAULT_MEASUREMENT_CHUNKS
+    kwargs: dict[str, Any] = {}
 
-    def all_measurement_groups(self) -> List[str]:
+    def all_measurement_groups(self) -> list[str]:
         ds, self.kwargs = open_dataset_autodetect(
             self.product_urlpath, check_files_exist=True, **self.kwargs
         )
@@ -100,31 +99,31 @@ class Sentinel1SarProduct(sarsen.GroundRangeSarProduct, sarsen.SlantRangeSarProd
             ds = xarray_sentinel.mosaic_slc_iw(ds)
         return ds
 
-    @property
+    @functools.cached_property
     def orbit(self) -> xr.Dataset:
         ds, self.kwargs = open_dataset_autodetect(
             self.product_urlpath, group=f"{self.measurement_group}/orbit", **self.kwargs
         )
-        return ds
+        return ds.compute()
 
-    @property
+    @functools.cached_property
     def gcp(self) -> xr.Dataset:
         ds, self.kwargs = open_dataset_autodetect(
             self.product_urlpath, group=f"{self.measurement_group}/gcp", **self.kwargs
         )
-        return ds
+        return ds.compute()
 
-    @property
+    @functools.cached_property
     def calibration(self) -> xr.Dataset:
         ds, self.kwargs = open_dataset_autodetect(
             self.product_urlpath,
             group=f"{self.measurement_group}/calibration",
             **self.kwargs,
         )
-        return ds
+        return ds.compute()
 
-    @property
-    def coordinate_conversion(self) -> Optional[xr.Dataset]:
+    @functools.cached_property
+    def coordinate_conversion(self) -> xr.Dataset | None:
         ds = None
         if self.product_type == "GRD":
             ds, self.kwargs = open_dataset_autodetect(
@@ -132,10 +131,11 @@ class Sentinel1SarProduct(sarsen.GroundRangeSarProduct, sarsen.SlantRangeSarProd
                 group=f"{self.measurement_group}/coordinate_conversion",
                 **self.kwargs,
             )
+            ds = ds.compute()
         return ds
 
-    @property
-    def azimuth_fm_rate(self) -> Optional[xr.Dataset]:
+    @functools.cached_property
+    def azimuth_fm_rate(self) -> xr.Dataset | None:
         ds = None
         if self.product_type == "SLC":
             ds, self.kwargs = open_dataset_autodetect(
@@ -143,10 +143,11 @@ class Sentinel1SarProduct(sarsen.GroundRangeSarProduct, sarsen.SlantRangeSarProd
                 group=f"{self.measurement_group}/azimuth_fm_rate",
                 **self.kwargs,
             )
+            ds = ds.compute()
         return ds
 
-    @property
-    def dc_estimate(self) -> Optional[xr.Dataset]:
+    @functools.cached_property
+    def dc_estimate(self) -> xr.Dataset | None:
         ds = None
         if self.product_type == "SLC":
             ds, self.kwargs = open_dataset_autodetect(
@@ -154,23 +155,36 @@ class Sentinel1SarProduct(sarsen.GroundRangeSarProduct, sarsen.SlantRangeSarProd
                 group=f"{self.measurement_group}/dc_estimate",
                 **self.kwargs,
             )
+            ds = ds.compute()
         return ds
 
+    # make class hashable to allow caching methods calls
+
+    def __hash__(self) -> int:
+        id = (
+            self.product_urlpath,
+            self.measurement_group,
+            self.measurement_chunks,
+        ) + tuple(repr(self.kwargs))
+        return hash(id)
+
     # SarProduct interaface
 
-    @property
+    @functools.cached_property
     def product_type(self) -> Any:
         prod_type = self.measurement.attrs["product_type"]
         assert isinstance(prod_type, str)
         return prod_type
 
-    def beta_nought(self) -> xr.DataArray:
+    @functools.cache
+    def beta_nought(self, persist: bool = False) -> xr.DataArray:
         measurement = self.measurement.data_vars["measurement"]
         beta_nought = xarray_sentinel.calibrate_intensity(
             measurement, self.calibration.betaNought
         )
-        beta_nought = beta_nought.drop_vars(["pixel", "line"])
-        return beta_nought
+        if persist:
+            beta_nought = beta_nought.persist()
+        return beta_nought.drop_vars(["pixel", "line"])
 
     def state_vectors(self) -> xr.DataArray:
         return self.orbit.data_vars["position"]
@@ -190,8 +204,8 @@ class Sentinel1SarProduct(sarsen.GroundRangeSarProduct, sarsen.SlantRangeSarProd
 
     def grid_parameters(
         self,
-        grouping_area_factor: Tuple[float, float] = (3.0, 3.0),
-    ) -> Dict[str, Any]:
+        grouping_area_factor: tuple[float, float] = (3.0, 3.0),
+    ) -> dict[str, Any]:
         return azimuth_slant_range_grid(self.measurement.attrs, grouping_area_factor)
 
     def complex_amplitude(self) -> xr.DataArray:
@@ -208,7 +222,8 @@ class Sentinel1SarProduct(sarsen.GroundRangeSarProduct, sarsen.SlantRangeSarProd
         else:
             return sarsen.SlantRangeSarProduct.interp_sar(self, *args, **kwargs)
 
-    def product_info(self, **kwargs: Any) -> Dict[str, Any]:
+    @functools.cache
+    def product_info(self, **kwargs: Any) -> dict[str, Any]:
         """Get information about the Sentinel-1 product."""
         measurement_groups = self.all_measurement_groups()
 
