@@ -1,4 +1,5 @@
 import numpy as np
+import pytest
 import xarray as xr
 
 from sarsen import scene
@@ -12,6 +13,7 @@ def test_convert_to_dem_3d(dem_raster: xr.DataArray) -> None:
     assert res.sel(x=12.5, y=42, method="nearest")[2] == 17.0
 
 
+@pytest.mark.xfail()
 def test_transform_dem_3d(dem_raster: xr.DataArray) -> None:
     dem_3d = scene.convert_to_dem_3d(dem_raster)
 
@@ -33,6 +35,13 @@ def test_transform_dem_3d(dem_raster: xr.DataArray) -> None:
     )
 
 
+def test_upsample(dem_raster: xr.DataArray) -> None:
+    res = scene.upsample(dem_raster, x=2)
+
+    assert res.x.size == dem_raster.x.size * 2
+    assert np.allclose(res.isel(x=0).mean(), 78.46527778)
+
+
 def test_compute_dem_oriented_area(dem_raster: xr.DataArray) -> None:
     dem_3d = scene.convert_to_dem_3d(dem_raster)
 
