@@ -1,3 +1,4 @@
+import pytest
 from typer.testing import CliRunner
 
 from sarsen import __main__
@@ -5,6 +6,7 @@ from sarsen import __main__
 runner = CliRunner()
 
 
+@pytest.mark.xfail()
 def test_main() -> None:
     res = runner.invoke(__main__.app, ["gtc", "--help"])
     assert res.exit_code == 0
